@@ -11,16 +11,14 @@ This allows us to catch exceptions at higher levels without forcing
 imports.
 """
 
+from __future__ import annotations
 
 import difflib
 
 from typing import (
     AnyStr,
     Iterable,
-    List,
-    Optional,
     Sequence,
-    Union,
 )
 
 # Do not import anything but pycompat here, please
@@ -44,8 +42,8 @@ class Hint:
     """
 
     def __init__(self, *args, **kw):
-        self.hint: Optional[bytes] = kw.pop('hint', None)
-        super(Hint, self).__init__(*args, **kw)
+        self.hint: bytes | None = kw.pop('hint', None)
+        super().__init__(*args, **kw)
 
 
 class Error(Hint, Exception):
@@ -54,7 +52,16 @@ class Error(Hint, Exception):
     coarse_exit_code = None
     detailed_exit_code = None
 
-    def __init__(self, message: bytes, hint: Optional[bytes] = None) -> None:
+    def __init__(
+        self,
+        message: bytes | str,
+        hint: bytes | None = None,
+    ) -> None:
+        # Ideally we would only accept "bytes" here, however the C code set
+        # error message as `str`, so we need to handle them on the fly or we
+        # will crash later down the road.
+        if isinstance(message, str):
+            message = pycompat.sysbytes(message)
         self.message = message
         self.hint = hint
         # Pass the message into the Exception constructor to help extensions
@@ -146,10 +153,10 @@ class ManifestLookupError(LookupError):
 class CommandError(Exception):
     """Exception raised on errors in parsing the command line."""
 
-    def __init__(self, command: Optional[bytes], message: bytes) -> None:
+    def __init__(self, command: bytes | None, message: bytes) -> None:
         self.command = command
         self.message = message
-        super(CommandError, self).__init__()
+        super().__init__()
 
     __bytes__ = _tobytes
 
@@ -160,11 +167,11 @@ class UnknownCommand(Exception):
     def __init__(
         self,
         command: bytes,
-        all_commands: Optional[List[bytes]] = None,
+        all_commands: list[bytes] | None = None,
     ) -> None:
         self.command = command
         self.all_commands = all_commands
-        super(UnknownCommand, self).__init__()
+        super().__init__()
 
     __bytes__ = _tobytes
 
@@ -172,10 +179,10 @@ class UnknownCommand(Exception):
 class AmbiguousCommand(Exception):
     """Exception raised if command shortcut matches more than one command."""
 
-    def __init__(self, prefix: bytes, matches: List[bytes]) -> None:
+    def __init__(self, prefix: bytes, matches: list[bytes]) -> None:
         self.prefix = prefix
         self.matches = matches
-        super(AmbiguousCommand, self).__init__()
+        super().__init__()
 
     __bytes__ = _tobytes
 
@@ -186,7 +193,7 @@ class WorkerError(Exception):
     def __init__(self, status_code: int) -> None:
         self.status_code = status_code
         # Pass status code to superclass just so it becomes part of __bytes__
-        super(WorkerError, self).__init__(status_code)
+        super().__init__(status_code)
 
     __bytes__ = _tobytes
 
@@ -281,10 +288,10 @@ class ConfigError(Abort):
     def __init__(
         self,
         message: bytes,
-        location: Optional[bytes] = None,
-        hint: Optional[bytes] = None,
+        location: bytes | None = None,
+        hint: bytes | None = None,
     ) -> None:
-        super(ConfigError, self).__init__(message, hint=hint)
+        super().__init__(message, hint=hint)
         self.location = location
 
     def format(self) -> bytes:
@@ -338,8 +345,8 @@ class OutOfBandError(RemoteError):
 
     def __init__(
         self,
-        message: Optional[bytes] = None,
-        hint: Optional[bytes] = None,
+        message: bytes | None = None,
+        hint: bytes | None = None,
     ):
         from .i18n import _
 
@@ -348,7 +355,7 @@ class OutOfBandError(RemoteError):
             message = _(b"remote error:\n%s") % message.rstrip(b'\n')
         else:
             message = _(b"remote error")
-        super(OutOfBandError, self).__init__(message, hint=hint)
+        super().__init__(message, hint=hint)
 
 
 class ParseError(Abort):
@@ -359,10 +366,10 @@ class ParseError(Abort):
     def __init__(
         self,
         message: bytes,
-        location: Optional[Union[bytes, int]] = None,
-        hint: Optional[bytes] = None,
+        location: bytes | int | None = None,
+        hint: bytes | None = None,
     ):
-        super(ParseError, self).__init__(message, hint=hint)
+        super().__init__(message, hint=hint)
         self.location = location
 
     def format(self) -> bytes:
@@ -392,14 +399,14 @@ class PatchApplicationError(PatchError):
     __bytes__ = _tobytes
 
 
-def getsimilar(symbols: Iterable[bytes], value: bytes) -> List[bytes]:
+def getsimilar(symbols: Iterable[bytes], value: bytes) -> list[bytes]:
     sim = lambda x: difflib.SequenceMatcher(None, value, x).ratio()
     # The cutoff for similarity here is pretty arbitrary. It should
     # probably be investigated and tweaked.
     return [s for s in symbols if sim(s) > 0.6]
 
 
-def similarity_hint(similar: List[bytes]) -> Optional[bytes]:
+def similarity_hint(similar: list[bytes]) -> bytes | None:
     from .i18n import _
 
     if len(similar) == 1:
@@ -476,11 +483,11 @@ class UnknownVersion(Abort):
     def __init__(
         self,
         msg: bytes,
-        hint: Optional[bytes] = None,
-        version: Optional[bytes] = None,
+        hint: bytes | None = None,
+        version: bytes | None = None,
     ) -> None:
         self.version = version
-        super(UnknownVersion, self).__init__(msg, hint=hint)
+        super().__init__(msg, hint=hint)
 
 
 class LockError(IOError):
@@ -489,7 +496,7 @@ class LockError(IOError):
         errno: int,
         strerror: str,
         filename: bytes,
-        desc: Optional[bytes],
+        desc: bytes | None,
     ) -> None:
         IOError.__init__(self, errno, strerror, filename)
         self.desc = desc
@@ -502,7 +509,7 @@ class LockHeld(LockError):
         self,
         errno: int,
         filename: bytes,
-        desc: Optional[bytes],
+        desc: bytes | None,
         locker,
     ):
         LockError.__init__(self, errno, 'Lock held', filename, desc)
@@ -548,7 +555,7 @@ class ProgrammingError(Hint, RuntimeError):
         # an internal-only error that won't be printed except in a
         # stack traces.
         msg = pycompat.sysstr(msg)
-        super(ProgrammingError, self).__init__(msg, *args, **kwargs)
+        super().__init__(msg, *args, **kwargs)
 
     __bytes__ = _tobytes
 
@@ -685,7 +692,7 @@ class WireprotoCommandError(Exception):
     def __init__(
         self,
         message: bytes,
-        args: Optional[Sequence[bytes]] = None,
+        args: Sequence[bytes] | None = None,
     ) -> None:
         self.message = message
         self.messageargs = args
